
import logging
import os
import sys
import platform
import time
import re
import multiprocessing
import socket
try:
    import resource
except ImportError:
    pass

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import parse_semver
from graphsignal import version

logger = logging.getLogger('graphsignal')

OS_LINUX = (sys.platform.startswith('linux'))
OS_DARWIN = (sys.platform == 'darwin')
OS_WIN = (sys.platform == 'win32')
CPU_NAME_REGEXP = re.compile(r'Model name:\s+(.+)$', flags=re.MULTILINE)
CPU_NAME_MAC_REGEXP = re.compile(r'machdep\.cpu\.brand_string:\s+(.+)$', flags=re.MULTILINE)
VM_RSS_REGEXP = re.compile(r'VmRSS:\s+(\d+)\s+kB')
VM_SIZE_REGEXP = re.compile(r'VmSize:\s+(\d+)\s+kB')
MEM_TOTAL_REGEXP = re.compile(r'MemTotal:\s+(\d+)\s+kB')
MEM_FREE_REGEXP = re.compile(r'MemFree:\s+(\d+)\s+kB')


class ProcessReader():
    MIN_CPU_READ_INTERVAL_NS = 1 * 1e9

    def __init__(self):
        self._start_ms = int(time.time() * 1e3)
        self._last_read_sec = None
        self._last_cpu_time_ns = None

    def setup(self):
        pass

    def shutdown(self):
        pass

    def read(self, signal):
        now = time.time()
        pid = str(os.getpid())

        node_usage = signal.node_usage
        process_usage = signal.process_usage
        process_usage.start_ms = self._start_ms

        process_usage.process_id = pid

        if not OS_WIN:
            cpu_time_ns = _read_cpu_time()
            if cpu_time_ns is not None:
                if self._last_cpu_time_ns is not None:
                    interval_ns = (now - self._last_read_sec) * 1e9
                    if interval_ns > ProcessReader.MIN_CPU_READ_INTERVAL_NS:
                        cpu_diff_ns = cpu_time_ns - self._last_cpu_time_ns
                        cpu_usage = (cpu_diff_ns / interval_ns) * 100
                        try:
                            cpu_usage = cpu_usage / multiprocessing.cpu_count()
                        except Exception:
                            pass
                        process_usage.cpu_usage_percent = cpu_usage

                if (self._last_read_sec is None or now - self._last_read_sec > 0):
                    self._last_read_sec = now
                    self._last_cpu_time_ns = cpu_time_ns

        if not OS_WIN:
            max_rss = _read_max_rss()
            if max_rss is not None:
                process_usage.max_rss = max_rss

        if OS_LINUX:
            current_rss = _read_current_rss()
            if current_rss is not None:
                process_usage.current_rss = current_rss

            vm_size = _read_vm_size()
            if vm_size is not None:
                process_usage.vm_size = vm_size

        try:
            node_usage.hostname = socket.gethostname()
            if node_usage.hostname:
                node_usage.ip_address = socket.gethostbyname(node_usage.hostname)
        except BaseException:
            logger.debug('Error reading hostname', exc_info=True)

        mem_total = _read_mem_total()
        if mem_total is not None:
            node_usage.mem_total = mem_total
            mem_free = _read_mem_free()
            if mem_free is not None:
                node_usage.mem_used = mem_total - mem_free

        try:
            node_usage.platform = sys.platform
            node_usage.machine = platform.machine()
            if not OS_WIN:
                node_usage.os_name = os.uname().sysname
                node_usage.os_version = os.uname().release
        except BaseException:
            logger.error('Error reading node information', exc_info=True)

        try:
            process_usage.runtime = signals_pb2.ProcessUsage.Runtime.PYTHON
            process_usage.runtime_version.major = sys.version_info.major
            process_usage.runtime_version.minor = sys.version_info.minor
            process_usage.runtime_version.patch = sys.version_info.micro
            process_usage.runtime_impl = platform.python_implementation()
            parse_semver(signal.agent_info.version, version.__version__)
        except BaseException:
            logger.error('Error reading process information', exc_info=True)


def _read_cpu_time():
    rusage = resource.getrusage(resource.RUSAGE_SELF)
    return int((rusage.ru_utime + rusage.ru_stime) * 1e9)  # ns


def _read_max_rss():
    rusage = resource.getrusage(resource.RUSAGE_SELF)

    if OS_DARWIN:
        return rusage.ru_maxrss
    else:
        return rusage.ru_maxrss * 1024


def _read_current_rss():
    pid = os.getpid()

    try:
        f = open('/proc/{0}/status'.format(os.getpid()))
        output = f.read()
        f.close()
    except Exception:
        return None

    match = VM_RSS_REGEXP.search(output)
    if match:
        return int(float(match.group(1)) * 1e3)

    return None


def _read_vm_size():
    pid = os.getpid()

    try:
        f = open('/proc/{0}/status'.format(os.getpid()))
        output = f.read()
        f.close()
    except BaseException:
        return None

    match = VM_SIZE_REGEXP.search(output)
    if match:
        return int(float(match.group(1)) * 1e3)

    return None

def _read_mem_total():
    try:
        f = open('/proc/meminfo')
        output = f.read()
        f.close()
    except BaseException:
        return None

    match = MEM_TOTAL_REGEXP.search(output)
    if match:
        return int(float(match.group(1)) * 1e3)

    return None

def _read_mem_free():
    try:
        f = open('/proc/meminfo')
        output = f.read()
        f.close()
    except BaseException:
        return None

    match = MEM_FREE_REGEXP.search(output)
    if match:
        return int(float(match.group(1)) * 1e3)

    return None



import logging
import threading
import random
import time
import uuid
import hashlib

import graphsignal
from graphsignal.tracer import Tracer
from graphsignal.uploader import Uploader
from graphsignal.usage.process_reader import ProcessReader
from graphsignal.usage.nvml_reader import NvmlReader
from graphsignal.trace_sampler import TraceSampler
from graphsignal.proto import signals_pb2
from graphsignal.data.missing_value_detector import MissingValueDetector

logger = logging.getLogger('graphsignal')


class Agent:
    def __init__(self, api_key, deployment=None, debug_mode=False):
        self.worker_id = _uuid_sha1(size=12)
        self.api_key = api_key
        self.deployment = deployment
        self.debug_mode = debug_mode

        self._uploader = None
        self._process_reader = None
        self._nvml_reader = None
        self._trace_samplers = None
        self._metric_store = None
        self._tracers = None
        self._mv_detectors = None

    def start(self):
        self._uploader = Uploader()
        self._uploader.setup()
        self._process_reader = ProcessReader()
        self._process_reader.setup()
        self._nvml_reader = NvmlReader()
        self._nvml_reader.setup()
        self._trace_samplers = {}
        self._metric_store = {}
        self._tracers = {}
        self._mv_detectors = {}

    def stop(self):
        self.upload(block=True)
        self._process_reader.shutdown()
        self._nvml_reader.shutdown()
        self._trace_samplers = None
        self._metric_store = None
        self._tracers = None

    def uploader(self):
        return self._uploader

    def tracer(self, with_profiler=True):
        if with_profiler in self._tracers:
            return self._tracers[with_profiler]

        profiler = None
        if with_profiler == True or with_profiler == 'python':
            from graphsignal.profilers.python import PythonProfiler
            profiler = PythonProfiler()
        elif with_profiler == 'tensorflow':
            from graphsignal.profilers.tensorflow import TensorFlowProfiler
            profiler = TensorFlowProfiler()
        elif with_profiler == 'pytorch':
            from graphsignal.profilers.pytorch import PyTorchProfiler
            profiler = PyTorchProfiler()
        elif with_profiler == 'jax':
            from graphsignal.profilers.jax import JaxProfiler
            profiler = JaxProfiler()
        elif with_profiler == 'onnxruntime':
            from graphsignal.profilers.onnxruntime import ONNXRuntimeProfiler
            profiler = ONNXRuntimeProfiler()
        elif with_profiler:
            raise ValueError('Invalid profiler name: {0}'.format(with_profiler))

        tracer = self._tracers[with_profiler] = Tracer(profiler=profiler)
        return tracer

    def get_trace_sampler(self, endpoint):
        if endpoint in self._trace_samplers:
            return self._trace_samplers[endpoint]
        else:
            trace_sampler = self._trace_samplers[endpoint] = TraceSampler()
            return trace_sampler

    def reset_metric_store(self, endpoint):
        self._metric_store[endpoint] = MetricStore()

    def get_metric_store(self, endpoint):
        if endpoint not in self._metric_store:
            self.reset_metric_store(endpoint)

        return self._metric_store[endpoint]

    def get_mv_detector(self, endpoint):
        if endpoint in self._mv_detectors:
            return self._mv_detectors[endpoint]
        else:
            mv_detector = self._mv_detectors[endpoint] = MissingValueDetector()
            return mv_detector

    def read_usage(self, signal):
        self._process_reader.read(signal)
        self._nvml_reader.read(signal)

    def create_signal(self):
        signal = signals_pb2.WorkerSignal()
        signal.worker_id = graphsignal._agent.worker_id
        signal.signal_id = _uuid_sha1(size=12)
        if self.deployment:
            signal.deployment_name = self.deployment
        return signal

    def upload(self, block=False):
        if block:
            self._uploader.flush()
        else:
            self._uploader.flush_in_thread()

    def tick(self, block=False):
        self.upload(block=False)


class MetricStore:
    MAX_RESERVOIR_SIZE = 100

    def __init__(self):
        self._update_lock = threading.Lock()
        self._start_sec = int(time.time())
        self.latency_us = None
        self.call_count = None
        self.exception_count = None
        self.data_counters = {}

    def add_time(self, duration_us):
        with self._update_lock:
            if not self.latency_us:
                self.latency_us = signals_pb2.Metric()
                self.latency_us.type = signals_pb2.Metric.MetricType.RESERVOIR_METRIC

            if len(self.latency_us.reservoir.values) < MetricStore.MAX_RESERVOIR_SIZE:
                self.latency_us.reservoir.values.append(duration_us)
            else:
                self.latency_us.reservoir.values[random.randint(0, MetricStore.MAX_RESERVOIR_SIZE - 1)] = duration_us

    def inc_call_count(self, value, timestamp_us):
        with self._update_lock:
            if not self.call_count:
                self.call_count = signals_pb2.Metric()
                self.call_count.type = signals_pb2.Metric.MetricType.COUNTER_METRIC
            
            self._inc_counter(self.call_count, value, timestamp_us)

    def inc_exception_count(self, value, timestamp_us):
        with self._update_lock:
            if not self.exception_count:
                self.exception_count = signals_pb2.Metric()
                self.exception_count.type = signals_pb2.Metric.MetricType.COUNTER_METRIC
            
            self._inc_counter(self.exception_count, value, timestamp_us)

    def inc_data_counter(self, data_name, counter_name, value, timestamp_us):
        with self._update_lock:
            key = (data_name, counter_name)
            if key not in self.data_counters:
                counter = signals_pb2.DataMetric()
                counter.data_name = data_name
                counter.metric_name = counter_name
                counter.metric.type = signals_pb2.Metric.MetricType.COUNTER_METRIC
                self.data_counters[key] = counter
            else:
                counter = self.data_counters[key]

            self._inc_counter(counter.metric, value, timestamp_us)

    def finalize(self, timestamp_us):
        with self._update_lock:
            if self.call_count:
                self._finalize_counter(self.call_count, timestamp_us)
            if self.exception_count:
                self._finalize_counter(self.exception_count, timestamp_us)
            for counter in self.data_counters.values():
                self._finalize_counter(counter.metric, timestamp_us)

    def _inc_counter(self, counter, value, timestamp_us):
        bucket = int(timestamp_us / 1e6)
        if bucket in counter.counter.buckets:
            counter.counter.buckets[bucket] += value
            counter.counter.buckets[self._start_sec] = 0
        else:
            counter.counter.buckets[self._start_sec] = 0
            counter.counter.buckets[bucket] = value

    def _finalize_counter(self, counter, timestamp_us):
        end_sec = int(timestamp_us / 1e6)
        if end_sec not in counter.counter.buckets:
            counter.counter.buckets[end_sec] = 0


def _sha1(text, size=-1):
    sha1_hash = hashlib.sha1()
    sha1_hash.update(text.encode('utf-8'))
    return sha1_hash.hexdigest()[0:size]


def _uuid_sha1(size=-1):
    return _sha1(str(uuid.uuid4()), size)



import logging
import importlib
import functools

import graphsignal
from graphsignal.data.builtin_types import BuiltInTypesProfiler
from graphsignal.data.numpy_ndarray import NumpyNDArrayProfiler
from graphsignal.data.tf_tensor import TFTensorProfiler
from graphsignal.data.torch_tensor import TorchTensorProfiler

logger = logging.getLogger('graphsignal')


data_profilers = [
    BuiltInTypesProfiler(),
    NumpyNDArrayProfiler(),
    TFTensorProfiler(),
    TorchTensorProfiler()
]


def compute_counts(data):
    for dp in data_profilers:
        if dp.is_instance(data):
            return dp.compute_counts(data)


def build_stats(data):
    for dp in data_profilers:
        if dp.is_instance(data):
            return dp.build_stats(data)



import logging

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data.data_profiler import DataProfiler, add_counts

logger = logging.getLogger('graphsignal')


class NumpyNDArrayProfiler(DataProfiler):
    def __init__(self):
        super().__init__()

    def is_instance(self, data):
        np = self.check_module('numpy')        
        return np is not None and isinstance(data, np.ndarray)

    def compute_counts(self, data):
        np = self.check_module('numpy')
        counts = {}
        counts['element_count'] = data.size
        counts['null_count'] = np.count_nonzero(data == None)
        if np.issubdtype(data.dtype, np.number):
            counts['nan_count'] = np.count_nonzero(np.isnan(data))
            counts['inf_count'] = np.count_nonzero(np.isinf(data))
        counts['zero_count'] = np.count_nonzero(data == 0)
        counts['negative_count'] = np.count_nonzero(data < 0)
        counts['positive_count'] = np.count_nonzero(data > 0)
        return counts

    def build_stats(self, data):
        counts = self.compute_counts(data)
        data_stats = signals_pb2.DataStats()
        data_stats.data_type = 'numpy.ndarray'
        data_stats.shape[:] = list(data.shape)
        add_counts(data_stats, counts)
        return data_stats


from typing import Optional
import logging
import os
import sys
import time
import gzip
import torch
from torch.autograd import DeviceType
import torch.distributed

import graphsignal
from graphsignal.proto_utils import parse_semver
from graphsignal.proto import signals_pb2
from graphsignal.profilers.operation_profiler import OperationProfiler
from graphsignal.profilers.profiler_utils import create_log_dir, remove_log_dir

logger = logging.getLogger('graphsignal')


class PyTorchProfiler(OperationProfiler):
    def __init__(self):
        self._torch_prof = None
        self._log_dir = None
        self._pytorch_version = None

    def read_info(self, signal):
        if not self._pytorch_version:
            self._pytorch_version = signals_pb2.SemVer()
            parse_semver(self._pytorch_version, torch.__version__)

        # Framework info
        framework = signal.frameworks.add()
        framework.type = signals_pb2.FrameworkInfo.FrameworkType.PYTORCH_FRAMEWORK
        framework.version.CopyFrom(self._pytorch_version)

    def start(self, signal):
        logger.debug('Activating PyTorch profiler')

        if not self._torch_prof:
            # Initialization
            def _schedule_func(step):
                return torch.profiler.ProfilerAction.RECORD

            self._torch_prof = torch.profiler.profile(
                schedule=_schedule_func,
                record_shapes=False,
                profile_memory=True,
                with_stack=False,
                with_flops=True)

        # Profiler info
        signal.agent_info.operation_profiler_type = signals_pb2.AgentInfo.ProfilerType.PYTORCH_PROFILER

        self._torch_prof.start()

    def stop(self, signal):
        logger.debug('Deactivating PyTorch profiler')

        self._torch_prof.stop()

        self._convert_operations(signal)

        self._read_chrome_trace(signal)

    def _convert_operations(self, signal):
        # Operation stats
        for event_avg in self._torch_prof.key_averages():
            if event_avg.key and event_avg.key.startswith('ProfilerStep'):
                continue
            op_stats = signal.op_stats.add()
            op_stats.op_name = event_avg.key
            op_stats.count = _uint(event_avg.count)
            op_stats.total_host_time_us = _uint(event_avg.cpu_time_total)
            op_stats.total_device_time_us = _uint(event_avg.cuda_time_total)
            op_stats.self_host_time_us = _uint(event_avg.self_cpu_time_total)
            op_stats.self_device_time_us = _uint(event_avg.self_cuda_time_total)
            op_stats.total_host_memory = _uint(event_avg.cpu_memory_usage)
            op_stats.total_device_memory = _uint(event_avg.cuda_memory_usage)
            op_stats.self_host_memory = _uint(event_avg.self_cpu_memory_usage)
            op_stats.self_device_memory = _uint(event_avg.self_cuda_memory_usage)
            op_stats.flops = _uint(event_avg.flops)
            if event_avg.device_type in (DeviceType.CUDA, DeviceType.HIP)  or op_stats.self_device_time_us > 0:
                op_stats.device_type = signals_pb2.DeviceType.GPU
            else:
                op_stats.device_type = signals_pb2.DeviceType.CPU

        # Kernel stats
        kernel_index = {}
        for event in self._torch_prof.events():
            for kernel in event.kernels:
                key = (event.key, kernel.name, kernel.device)
                if key in kernel_index:
                    kernel_stats = kernel_index[key]
                    kernel_stats.count += 1
                    kernel_stats.duration_ns += _uint(kernel.duration * 1000)
                else:
                    kernel_stats = kernel_index[key] = signal.kernel_stats.add()
                    kernel_stats.device_type = signals_pb2.DeviceType.GPU
                    kernel_stats.device_id = str(kernel.device)
                    kernel_stats.op_name = event.name
                    kernel_stats.kernel_name = kernel.name
                    kernel_stats.count = 1
                    kernel_stats.duration_ns = _uint(kernel.duration * 1000)

        for kernel_stats in kernel_index.values():
            signal.kernel_stats.append(kernel_stats)

        logger.debug(
            'Converted %d PyTorch operation statistics', len(signal.op_stats))

    def _read_chrome_trace(self, signal):
        try:
            self._log_dir = create_log_dir()

            trace_path = os.path.join(self._log_dir, 'trace.json')
            self._torch_prof.export_chrome_trace(trace_path)

            trace_file_size = os.path.getsize(trace_path)
            if trace_file_size > 50 * 1e6:
                raise Exception('Trace file too big: {0}'.format(trace_file_size))

            with open(trace_path) as f:
                trace_json = f.read()
                signal.trace_data = gzip.compress(trace_json.encode())
        finally:
            remove_log_dir(self._log_dir)

def _uint(val):
    return max(int(val), 0)



import unittest
import logging
import sys
import time
import numpy as np
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.tracer import EndpointTrace
from graphsignal.profilers.tensorflow import TensorFlowProfiler
from graphsignal.usage.process_reader import ProcessReader
from graphsignal.usage.nvml_reader import NvmlReader
from graphsignal.uploader import Uploader
from graphsignal.data.missing_value_detector import MissingValueDetector

logger = logging.getLogger('graphsignal')


class TracerTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='s1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    @patch.object(TensorFlowProfiler, 'start', return_value=True)
    @patch.object(TensorFlowProfiler, 'stop', return_value=True)
    @patch.object(ProcessReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_signal')
    @patch('time.time', return_value=1000)
    def test_start_stop(self, mocked_time, mocked_upload_signal, mocked_nvml_read, mocked_host_read,
                        mocked_stop, mocked_start):
        for i in range(10):
            trace = EndpointTrace(
                endpoint='ep1',
                tags={'k1': 'v2', 'k3': 3.0},
                operation_profiler=TensorFlowProfiler())
            trace.set_tag('k4', 'v4')
            trace.set_data('input', np.asarray([[1, 2],[3, 4]]))
            time.sleep(0.01)
            trace.stop()

        self.assertEqual(mocked_start.call_count, 2)
        self.assertEqual(mocked_stop.call_count, 2)
        signal = mocked_upload_signal.call_args[0][0]

        self.assertEqual(signal.deployment_name, 's1')
        self.assertEqual(signal.endpoint_name, 'ep1')
        self.assertTrue(signal.worker_id != '')
        self.assertTrue(signal.start_us > 0)
        self.assertTrue(signal.end_us > 0)
        self.assertEqual(signal.signal_type, signals_pb2.SignalType.PROFILE_SIGNAL)
        self.assertEqual(len(signal.trace_metrics.call_count.counter.buckets), 1)
        self.assertEqual(len(signal.trace_metrics.latency_us.reservoir.values), 8)
        self.assertEqual(signal.data_metrics[0].data_name, 'input')
        self.assertEqual(signal.data_metrics[0].metric_name, 'element_count')
        self.assertEqual(len(signal.data_metrics[0].metric.counter.buckets), 1)
        self.assertEqual(signal.tags[0].key, 'k1')
        self.assertEqual(signal.tags[0].value, 'v2')
        self.assertEqual(signal.tags[1].key, 'k3')
        self.assertEqual(signal.tags[1].value, '3.0')
        self.assertEqual(signal.tags[2].key, 'k4')
        self.assertEqual(signal.tags[2].value, 'v4')
        self.assertEqual(signal.tags[2].key, 'k4')
        self.assertEqual(signal.tags[2].value, 'v4')
        self.assertEqual(signal.data_stats[0].data_name, 'input')
        self.assertEqual(signal.data_stats[0].shape, [2, 2])

    @patch.object(TensorFlowProfiler, 'start', return_value=True)
    @patch.object(TensorFlowProfiler, 'stop', return_value=True)
    @patch.object(ProcessReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_signal')
    def test_start_exception(self, mocked_upload_signal, mocked_nvml_read, mocked_host_read,
                             mocked_stop, mocked_start):
        mocked_start.side_effect = Exception('ex1')
        trace = EndpointTrace(
            endpoint='ep1',
            operation_profiler=TensorFlowProfiler())
        trace.stop()

        mocked_start.assert_called_once()
        mocked_stop.assert_not_called()
        signal = mocked_upload_signal.call_args[0][0]

        self.assertEqual(signal.endpoint_name, 'ep1')
        self.assertTrue(signal.worker_id != '')
        self.assertTrue(signal.start_us > 0)
        self.assertTrue(signal.end_us > 0)
        self.assertEqual(signal.signal_type, signals_pb2.SignalType.SAMPLE_SIGNAL)
        self.assertEqual(signal.profiler_errors[0].message, 'ex1')
        self.assertNotEqual(signal.profiler_errors[0].stack_trace, '')

    @patch.object(TensorFlowProfiler, 'start', return_value=True)
    @patch.object(TensorFlowProfiler, 'stop', return_value=True)
    @patch.object(ProcessReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_signal')
    def test_profiler_exception(self, mocked_upload_signal, mocked_nvml_read, mocked_host_read,
                            mocked_stop, mocked_start):
        mocked_stop.side_effect = Exception('ex1')
        trace = EndpointTrace(
            endpoint='ep1',
            operation_profiler=TensorFlowProfiler())
        trace.stop()

        mocked_start.assert_called_once()
        mocked_stop.assert_called_once()
        signal = mocked_upload_signal.call_args[0][0]

        self.assertEqual(signal.endpoint_name, 'ep1')
        self.assertTrue(signal.worker_id != '')
        self.assertTrue(signal.start_us > 0)
        self.assertTrue(signal.end_us > 0)
        self.assertEqual(signal.profiler_errors[0].message, 'ex1')
        self.assertNotEqual(signal.profiler_errors[0].stack_trace, '')

    @patch.object(TensorFlowProfiler, 'start', return_value=True)
    @patch.object(TensorFlowProfiler, 'stop', return_value=True)
    @patch.object(ProcessReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_signal')
    def test_inference_exception(self, mocked_upload_signal, mocked_nvml_read, mocked_host_read,
                            mocked_stop, mocked_start):

        for _ in range(2):
            try:
                with EndpointTrace(endpoint='ep1', operation_profiler=TensorFlowProfiler()):
                    raise Exception('ex1')
            except Exception as ex:
                if str(ex) != 'ex1':
                    raise ex

        self.assertEqual(mocked_upload_signal.call_count, 2)
        signal = mocked_upload_signal.call_args[0][0]

        self.assertEqual(signal.endpoint_name, 'ep1')
        self.assertTrue(signal.worker_id != '')
        self.assertTrue(signal.signal_id != '')
        self.assertTrue(signal.start_us > 0)
        self.assertTrue(signal.end_us > 0)
        self.assertEqual(signal.signal_type, signals_pb2.SignalType.EXCEPTION_SIGNAL)
        self.assertEqual(len(signal.trace_metrics.exception_count.counter.buckets), 1)
        self.assertEqual(signal.exceptions[0].exc_type, 'Exception')
        self.assertEqual(signal.exceptions[0].message, 'ex1')
        self.assertNotEqual(signal.exceptions[0].stack_trace, '')

    @patch.object(TensorFlowProfiler, 'start', return_value=True)
    @patch.object(TensorFlowProfiler, 'stop', return_value=True)
    @patch.object(ProcessReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_signal')
    def test_set_exception(self, mocked_upload_signal, mocked_nvml_read, mocked_host_read,
                            mocked_stop, mocked_start):
        trace = EndpointTrace(endpoint='ep1', operation_profiler=TensorFlowProfiler())
        try:
            raise Exception('ex2')
        except Exception as ex:
            trace.set_exception(ex)
        trace.stop()

        signal = mocked_upload_signal.call_args[0][0]

        self.assertEqual(signal.endpoint_name, 'ep1')
        self.assertTrue(signal.worker_id != '')
        self.assertTrue(signal.start_us > 0)
        self.assertTrue(signal.end_us > 0)
        self.assertEqual(signal.signal_type, signals_pb2.SignalType.EXCEPTION_SIGNAL)
        self.assertEqual(len(signal.trace_metrics.exception_count.counter.buckets), 1)
        self.assertEqual(signal.exceptions[0].exc_type, 'Exception')
        self.assertEqual(signal.exceptions[0].message, 'ex2')
        self.assertNotEqual(signal.exceptions[0].stack_trace, '')

    @patch.object(TensorFlowProfiler, 'start', return_value=True)
    @patch.object(TensorFlowProfiler, 'stop', return_value=True)
    @patch.object(ProcessReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_signal')
    def test_set_exception_true(self, mocked_upload_signal, mocked_nvml_read, mocked_host_read,
                            mocked_stop, mocked_start):
        trace = EndpointTrace(endpoint='ep1', operation_profiler=TensorFlowProfiler())
        try:
            raise Exception('ex2')
        except Exception as ex:
            trace.set_exception(exc_info=True)
        trace.stop()

        signal = mocked_upload_signal.call_args[0][0]

        self.assertEqual(signal.endpoint_name, 'ep1')
        self.assertTrue(signal.worker_id != '')
        self.assertTrue(signal.start_us > 0)
        self.assertTrue(signal.end_us > 0)
        self.assertEqual(signal.signal_type, signals_pb2.SignalType.EXCEPTION_SIGNAL)
        self.assertEqual(len(signal.trace_metrics.exception_count.counter.buckets), 1)
        self.assertEqual(signal.exceptions[0].exc_type, 'Exception')
        self.assertEqual(signal.exceptions[0].message, 'ex2')
        self.assertNotEqual(signal.exceptions[0].stack_trace, '')

    @patch.object(TensorFlowProfiler, 'start', return_value=True)
    @patch.object(TensorFlowProfiler, 'stop', return_value=True)
    @patch.object(ProcessReader, 'read')
    @patch.object(NvmlReader, 'read')
    @patch.object(Uploader, 'upload_signal')
    def test_set_data(self, mocked_upload_signal, mocked_nvml_read, mocked_host_read,
                            mocked_stop, mocked_start):
        
        for i in range(MissingValueDetector.MIN_BASELINE_SIZE):
            with EndpointTrace(endpoint='ep1', operation_profiler=TensorFlowProfiler()) as trace:
                trace.set_data('d1', {'c1': i, 'c2': 1})
        with EndpointTrace(endpoint='ep1', operation_profiler=TensorFlowProfiler()) as trace:
            trace.set_data('d1', {'c1': 100, 'c2': None})

        signal = mocked_upload_signal.call_args[0][0]

        self.assertEqual(signal.endpoint_name, 'ep1')
        self.assertTrue(signal.worker_id != '')
        self.assertTrue(signal.start_us > 0)
        self.assertTrue(signal.end_us > 0)
        self.assertEqual(signal.signal_type, signals_pb2.SignalType.MISSING_VALUES_SIGNAL)



import unittest
import logging
import sys
import os
from unittest.mock import patch, Mock
import torch
import onnxruntime
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.profilers.onnxruntime import ONNXRuntimeProfiler
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')

TEST_MODEL_PATH = '/tmp/test_model.onnx'

class ONNXRuntimeProfilerTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
        graphsignal.shutdown()
        if os.path.exists(TEST_MODEL_PATH):
            os.remove(TEST_MODEL_PATH)

    def test_read_info(self):
        profiler = ONNXRuntimeProfiler()
        signal = signals_pb2.WorkerSignal()
        profiler.read_info(signal)

        self.assertEqual(
            signal.frameworks[0].type,
            signals_pb2.FrameworkInfo.FrameworkType.ONNX_FRAMEWORK)

    def test_start_stop(self):
        x = torch.arange(-5, 5, 0.1).view(-1, 1)
        y = -5 * x + 0.1 * torch.randn(x.size())
        model = torch.nn.Linear(1, 1)
        if torch.cuda.is_available():
            x = x.to('cuda:0')
            y = y.to('cuda:0')
            model = model.to('cuda:0')
        y1 = model(x)

        input_names = [ "input" ]
        output_names = [ "output" ]
        torch.onnx.export(model, x, TEST_MODEL_PATH, verbose=True, input_names=input_names, output_names=output_names)

        profiler = ONNXRuntimeProfiler()
        signal = signals_pb2.WorkerSignal()

        sess_options = onnxruntime.SessionOptions()
        sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL

        profiler.initialize_options(sess_options)
        session = onnxruntime.InferenceSession(TEST_MODEL_PATH, sess_options)

        profiler.set_onnx_session(session)

        profiler.start(signal)
        session.run(None, { 'input': x.detach().cpu().numpy() })
        profiler.stop(signal)

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(signal))

        self.assertNotEqual(signal.trace_data, b'')



from typing import Optional
import logging
import time
from torch import Tensor
import pytorch_lightning
from pytorch_lightning.callbacks.base import Callback

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import parse_semver

logger = logging.getLogger('graphsignal')


class GraphsignalCallback(Callback):
    def __init__(self, endpoint: str, tags: Optional[dict] = None):
        super().__init__()
        self._pl_version = None
        self._trace = None
        self._endpoint = endpoint
        self._tags = tags
        self._model_size_mb = None

    def on_validate_start(self, trainer, pl_module):
        self._configure_profiler(trainer, pl_module)

    def on_validate_end(self, trainer, pl_module):
        graphsignal.upload()

    def on_validate_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        self._start_profiler(trainer)

    def on_validate_batch_end(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        self._stop_profiler(trainer)

    def on_test_start(self, trainer, pl_module):
        self._configure_profiler(trainer, pl_module)

    def on_test_end(self, trainer, pl_module):
        graphsignal.upload()

    def on_test_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        self._start_profiler(trainer)

    def on_test_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
        self._stop_profiler(trainer)

    def on_predict_start(self, trainer, pl_module):
        self._configure_profiler(trainer, pl_module)

    def on_predict_end(self, trainer, pl_module):
        graphsignal.upload()

    def on_predict_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        self._start_profiler(trainer)

    def on_predict_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
        self._stop_profiler(trainer)

    def _configure_profiler(self, trainer, pl_module):
        try:
            self._pl_version = signals_pb2.SemVer()
            parse_semver(self._pl_version, pytorch_lightning.__version__)

            model_size_mb = pytorch_lightning.utilities.memory.get_model_size_mb(pl_module)
            if model_size_mb:
                self._model_size_mb = model_size_mb
        except Exception:
            logger.error('Error configuring PyTorch Lightning profiler', exc_info=True)

    def _start_profiler(self, trainer):
        if not self._trace:
            self._trace = graphsignal.tracer(with_profiler='pytorch').trace(
                endpoint=self._endpoint,
                tags=self._tags)

    def _stop_profiler(self, trainer):
        if self._trace:
            if self._trace.is_tracing():
                self._update_profile(trainer)
            self._trace.stop()
            self._trace = None

    def _update_profile(self, trainer):
        try:
            signal = self._trace._signal

            signal.agent_info.framework_profiler_type = signals_pb2.AgentInfo.ProfilerType.PYTORCH_LIGHTNING_PROFILER

            framework = signal.frameworks.add()
            framework.type = signals_pb2.FrameworkInfo.FrameworkType.PYTORCH_LIGHTNING_FRAMEWORK
            framework.version.CopyFrom(self._pl_version)

            signal.model_info.model_format = signals_pb2.ModelInfo.ModelFormat.PYTORCH_FORMAT
            if self._model_size_mb:
                signal.model_info.model_size_bytes = int(self._model_size_mb * 1e6)
        except Exception as exc:
            self._trace._add_profiler_exception(exc)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: graphsignal/profilers/tensorflow_proto/input_pipeline.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from graphsignal.profilers.tensorflow_proto import diagnostics_pb2 as graphsignal_dot_profilers_dot_tensorflow__proto_dot_diagnostics__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n;graphsignal/profilers/tensorflow_proto/input_pipeline.proto\x12\x13tensorflow.profiler\x1a\x19google/protobuf/any.proto\x1a\x38graphsignal/profilers/tensorflow_proto/diagnostics.proto\"\x80\x03\n\x12\x42ottleneckAnalysis\x12\x15\n\rinput_percent\x18\x07 \x01(\x01\x12\x16\n\x0eoutput_percent\x18\x08 \x01(\x01\x12\x14\n\x0cidle_percent\x18\t \x01(\x01\x12\x17\n\x0f\x63ompute_percent\x18\n \x01(\x01\x12\x1c\n\x14input_classification\x18\x01 \x01(\t\x12\x17\n\x0finput_statement\x18\x02 \x01(\t\x12$\n\x1ckernel_launch_classification\x18\x03 \x01(\t\x12\x1f\n\x17kernel_launch_statement\x18\x04 \x01(\t\x12 \n\x18\x61ll_other_classification\x18\x05 \x01(\t\x12\x1b\n\x13\x61ll_other_statement\x18\x06 \x01(\t\x12)\n!device_collectives_classification\x18\x0b \x01(\t\x12$\n\x1c\x64\x65vice_collectives_statement\x18\x0c \x01(\t\"\\\n\x0bStepSummary\x12\x0f\n\x07\x61verage\x18\x01 \x01(\x01\x12\x1a\n\x12standard_deviation\x18\x02 \x01(\x01\x12\x0f\n\x07minimum\x18\x03 \x01(\x01\x12\x0f\n\x07maximum\x18\x04 \x01(\x01\"\xe0\x02\n\x15PerGenericStepDetails\x12\x13\n\x0bstep_number\x18\x01 \x01(\x05\x12\x11\n\tstep_name\x18\x0e \x01(\t\x12\x14\n\x0cstep_time_ms\x18\x02 \x01(\x01\x12\x17\n\x0funknown_time_ms\x18\x03 \x01(\x01\x12\x1a\n\x12host_wait_input_ms\x18\x0b \x01(\x01\x12\x19\n\x11host_to_device_ms\x18\x0c \x01(\x01\x12\x11\n\toutput_ms\x18\x05 \x01(\x01\x12\x19\n\x11\x64\x65vice_compute_ms\x18\x06 \x01(\x01\x12\x1b\n\x13\x64\x65vice_to_device_ms\x18\x07 \x01(\x01\x12\x1d\n\x15\x64\x65vice_collectives_ms\x18\r \x01(\x01\x12\x17\n\x0fhost_compute_ms\x18\x08 \x01(\x01\x12\x17\n\x0fhost_prepare_ms\x18\t \x01(\x01\x12\x17\n\x0fhost_compile_ms\x18\n \x01(\x01J\x04\x08\x04\x10\x05\"\xa5\x01\n\x12InputTimeBreakdown\x12\x1d\n\x15\x64\x65manded_file_read_us\x18\x01 \x01(\x01\x12\x1d\n\x15\x61\x64vanced_file_read_us\x18\x02 \x01(\x01\x12\x18\n\x10preprocessing_us\x18\x03 \x01(\x01\x12\x12\n\nenqueue_us\x18\x04 \x01(\x01\x12#\n\x1bunclassified_non_enqueue_us\x18\x05 \x01(\x01\"\xa6\x01\n\x0eInputOpDetails\x12\x0f\n\x07op_name\x18\x01 \x01(\t\x12\r\n\x05\x63ount\x18\x02 \x01(\x04\x12\x12\n\ntime_in_ms\x18\x03 \x01(\x01\x12\x17\n\x0ftime_in_percent\x18\x04 \x01(\x01\x12\x17\n\x0fself_time_in_ms\x18\x05 \x01(\x01\x12\x1c\n\x14self_time_in_percent\x18\x06 \x01(\x01\x12\x10\n\x08\x63\x61tegory\x18\x07 \x01(\t\"\x84\x01\n#InputPipelineAnalysisRecommendation\x12\x0f\n\x07\x64\x65tails\x18\x01 \x03(\t\x12\x31\n\x13\x62ottleneck_analysis\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\x12\x19\n\x11summary_next_step\x18\x03 \x01(\t\"\x85\x06\n\x18GenericStepTimeBreakdown\x12\x41\n\x17unknown_time_ms_summary\x18\x01 \x01(\x0b\x32 .tensorflow.profiler.StepSummary\x12\x44\n\x1ahost_wait_input_ms_summary\x18\t \x01(\x0b\x32 .tensorflow.profiler.StepSummary\x12\x43\n\x19host_to_device_ms_summary\x18\n \x01(\x0b\x32 .tensorflow.profiler.StepSummary\x12:\n\x10input_ms_summary\x18\x0b \x01(\x0b\x32 .tensorflow.profiler.StepSummary\x12;\n\x11output_ms_summary\x18\x03 \x01(\x0b\x32 .tensorflow.profiler.StepSummary\x12\x43\n\x19\x64\x65vice_compute_ms_summary\x18\x04 \x01(\x0b\x32 .tensorflow.profiler.StepSummary\x12\x45\n\x1b\x64\x65vice_to_device_ms_summary\x18\x05 \x01(\x0b\x32 .tensorflow.profiler.StepSummary\x12G\n\x1d\x64\x65vice_collectives_ms_summary\x18\x0c \x01(\x0b\x32 .tensorflow.profiler.StepSummary\x12\x41\n\x17host_compute_ms_summary\x18\x06 \x01(\x0b\x32 .tensorflow.profiler.StepSummary\x12\x41\n\x17host_prepare_ms_summary\x18\x07 \x01(\x0b\x32 .tensorflow.profiler.StepSummary\x12\x41\n\x17host_compile_ms_summary\x18\x08 \x01(\x0b\x32 .tensorflow.profiler.StepSummaryJ\x04\x08\x02\x10\x03\"\x8a\x05\n\x1bInputPipelineAnalysisResult\x12\x15\n\rhardware_type\x18\t \x01(\t\x12;\n\x11step_time_summary\x18\x02 \x01(\x0b\x32 .tensorflow.profiler.StepSummary\x12?\n\x15input_percent_summary\x18\x03 \x01(\x0b\x32 .tensorflow.profiler.StepSummary\x12\x15\n\rinput_percent\x18\x0b \x01(\x01\x12\x16\n\x0eoutput_percent\x18\r \x01(\x01\x12\x14\n\x0cidle_percent\x18\x0e \x01(\x01\x12\x17\n\x0f\x63ompute_percent\x18\x0f \x01(\x01\x12*\n\x0cstep_details\x18\x04 \x03(\x0b\x32\x14.google.protobuf.Any\x12\x45\n\x14input_time_breakdown\x18\x05 \x01(\x0b\x32\'.tensorflow.profiler.InputTimeBreakdown\x12=\n\x10input_op_details\x18\x06 \x03(\x0b\x32#.tensorflow.profiler.InputOpDetails\x12P\n\x0erecommendation\x18\x07 \x01(\x0b\x32\x38.tensorflow.profiler.InputPipelineAnalysisRecommendation\x12\x31\n\x13step_time_breakdown\x18\x08 \x01(\x0b\x32\x14.google.protobuf.Any\x12\x35\n\x0b\x64iagnostics\x18\x0c \x01(\x0b\x32 .tensorflow.profiler.DiagnosticsJ\x04\x08\x01\x10\x02J\x04\x08\n\x10\x0b\x62\x06proto3')



_BOTTLENECKANALYSIS = DESCRIPTOR.message_types_by_name['BottleneckAnalysis']
_STEPSUMMARY = DESCRIPTOR.message_types_by_name['StepSummary']
_PERGENERICSTEPDETAILS = DESCRIPTOR.message_types_by_name['PerGenericStepDetails']
_INPUTTIMEBREAKDOWN = DESCRIPTOR.message_types_by_name['InputTimeBreakdown']
_INPUTOPDETAILS = DESCRIPTOR.message_types_by_name['InputOpDetails']
_INPUTPIPELINEANALYSISRECOMMENDATION = DESCRIPTOR.message_types_by_name['InputPipelineAnalysisRecommendation']
_GENERICSTEPTIMEBREAKDOWN = DESCRIPTOR.message_types_by_name['GenericStepTimeBreakdown']
_INPUTPIPELINEANALYSISRESULT = DESCRIPTOR.message_types_by_name['InputPipelineAnalysisResult']
BottleneckAnalysis = _reflection.GeneratedProtocolMessageType('BottleneckAnalysis', (_message.Message,), {
  'DESCRIPTOR' : _BOTTLENECKANALYSIS,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.input_pipeline_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.BottleneckAnalysis)
  })
_sym_db.RegisterMessage(BottleneckAnalysis)

StepSummary = _reflection.GeneratedProtocolMessageType('StepSummary', (_message.Message,), {
  'DESCRIPTOR' : _STEPSUMMARY,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.input_pipeline_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.StepSummary)
  })
_sym_db.RegisterMessage(StepSummary)

PerGenericStepDetails = _reflection.GeneratedProtocolMessageType('PerGenericStepDetails', (_message.Message,), {
  'DESCRIPTOR' : _PERGENERICSTEPDETAILS,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.input_pipeline_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.PerGenericStepDetails)
  })
_sym_db.RegisterMessage(PerGenericStepDetails)

InputTimeBreakdown = _reflection.GeneratedProtocolMessageType('InputTimeBreakdown', (_message.Message,), {
  'DESCRIPTOR' : _INPUTTIMEBREAKDOWN,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.input_pipeline_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.InputTimeBreakdown)
  })
_sym_db.RegisterMessage(InputTimeBreakdown)

InputOpDetails = _reflection.GeneratedProtocolMessageType('InputOpDetails', (_message.Message,), {
  'DESCRIPTOR' : _INPUTOPDETAILS,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.input_pipeline_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.InputOpDetails)
  })
_sym_db.RegisterMessage(InputOpDetails)

InputPipelineAnalysisRecommendation = _reflection.GeneratedProtocolMessageType('InputPipelineAnalysisRecommendation', (_message.Message,), {
  'DESCRIPTOR' : _INPUTPIPELINEANALYSISRECOMMENDATION,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.input_pipeline_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.InputPipelineAnalysisRecommendation)
  })
_sym_db.RegisterMessage(InputPipelineAnalysisRecommendation)

GenericStepTimeBreakdown = _reflection.GeneratedProtocolMessageType('GenericStepTimeBreakdown', (_message.Message,), {
  'DESCRIPTOR' : _GENERICSTEPTIMEBREAKDOWN,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.input_pipeline_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.GenericStepTimeBreakdown)
  })
_sym_db.RegisterMessage(GenericStepTimeBreakdown)

InputPipelineAnalysisResult = _reflection.GeneratedProtocolMessageType('InputPipelineAnalysisResult', (_message.Message,), {
  'DESCRIPTOR' : _INPUTPIPELINEANALYSISRESULT,
  '__module__' : 'graphsignal.profilers.tensorflow_proto.input_pipeline_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.profiler.InputPipelineAnalysisResult)
  })
_sym_db.RegisterMessage(InputPipelineAnalysisResult)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _BOTTLENECKANALYSIS._serialized_start=170
  _BOTTLENECKANALYSIS._serialized_end=554
  _STEPSUMMARY._serialized_start=556
  _STEPSUMMARY._serialized_end=648
  _PERGENERICSTEPDETAILS._serialized_start=651
  _PERGENERICSTEPDETAILS._serialized_end=1003
  _INPUTTIMEBREAKDOWN._serialized_start=1006
  _INPUTTIMEBREAKDOWN._serialized_end=1171
  _INPUTOPDETAILS._serialized_start=1174
  _INPUTOPDETAILS._serialized_end=1340
  _INPUTPIPELINEANALYSISRECOMMENDATION._serialized_start=1343
  _INPUTPIPELINEANALYSISRECOMMENDATION._serialized_end=1475
  _GENERICSTEPTIMEBREAKDOWN._serialized_start=1478
  _GENERICSTEPTIMEBREAKDOWN._serialized_end=2251
  _INPUTPIPELINEANALYSISRESULT._serialized_start=2254
  _INPUTPIPELINEANALYSISRESULT._serialized_end=2904
# @@protoc_insertion_point(module_scope)



import unittest
import logging
import sys
import os
import time
from unittest.mock import patch, Mock

import graphsignal

logger = logging.getLogger('graphsignal')


class AgentTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_tracer_default(self):
        tracer = graphsignal._agent.tracer()
        self.assertIsNotNone(tracer)
        self.assertTrue(isinstance(tracer.profiler(), graphsignal.profilers.python.PythonProfiler))

    def test_tracer_python(self):
        tracer = graphsignal._agent.tracer(with_profiler='python')
        self.assertIsNotNone(tracer)
        self.assertTrue(isinstance(tracer.profiler(), graphsignal.profilers.python.PythonProfiler))

    def test_tracer_disabled(self):
        tracer = graphsignal._agent.tracer(with_profiler=False)
        self.assertIsNotNone(tracer)
        self.assertIsNone(tracer.profiler())

    @patch('time.time', return_value=1)
    def test_update_metric_store(self, mocked_time):
        store = graphsignal._agent.get_metric_store('ep1')

        store.inc_call_count(1, 1000 * 1e6)
        store.inc_call_count(1, 1000 * 1e6)
        store.inc_call_count(1, 1001 * 1e6)
        self.assertEqual(store.call_count.counter.buckets, {1: 0, 1000: 2, 1001: 1})
        
        store.inc_exception_count(1, 1003 * 1e6)
        store.inc_exception_count(1, 1003 * 1e6)
        store.inc_exception_count(1, 1004 * 1e6)
        self.assertEqual(store.exception_count.counter.buckets, {1: 0, 1003: 2, 1004: 1})

        store.inc_data_counter('d1', 'c1', 1, 1000 * 1e6)
        store.inc_data_counter('d1', 'c1', 1, 1000 * 1e6)
        store.inc_data_counter('d1', 'c1', 1, 1001 * 1e6)
        store.inc_data_counter('d1', 'c2', 1, 1001 * 1e6)
        self.assertEqual(store.data_counters[('d1', 'c1')].metric.counter.buckets, {1: 0, 1000: 2, 1001: 1})
        self.assertEqual(store.data_counters[('d1', 'c2')].metric.counter.buckets, {1: 0, 1001: 1})

        store.add_time(20)
        store.add_time(30)
        self.assertEqual(store.latency_us.reservoir.values, [20, 30])




import logging
import time
import random
from threading import Lock

logger = logging.getLogger('graphsignal')

# global sampling lock
_sampling_lock = Lock()

class TraceSampler:
    MIN_INTERVAL_SEC = 60

    def __init__(self):
        self._current_trace_idx = 0
        self._trace_counters = {}
        self._last_reset_ts = time.time()

    def current_trace_idx(self):
        return self._current_trace_idx

    def lock(self, group, limit_per_interval=1, include_trace_idx=None):
        self._current_trace_idx += 1

        if _sampling_lock.locked():
            return False

        now = time.time()

        # reset counters
        if now - self._last_reset_ts > self.MIN_INTERVAL_SEC:
            self._trace_counters = {}
            self._last_reset_ts = time.time()

        # check counters
        if group in self._trace_counters:
            self._trace_counters[group] += 1
        else:
            self._trace_counters[group] = 1
        if self._trace_counters[group] <= limit_per_interval:
            _sampling_lock.acquire(blocking=False)
            return True

        # check include_trace_idx
        if include_trace_idx is not None and self._current_trace_idx in include_trace_idx:
            _sampling_lock.acquire(blocking=False)
            return True

        return False

    def unlock(self):
        if not _sampling_lock.locked():
            return
        _sampling_lock.release()



import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data.missing_value_detector import MissingValueDetector

logger = logging.getLogger('graphsignal')


class MissingValueDetectorTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_build_stats_none(self):
        mvd = MissingValueDetector()
        for i in range(MissingValueDetector.MIN_BASELINE_SIZE):
            self.assertFalse(mvd.detect('d1', {'null_count': 0}))
        self.assertTrue(mvd.detect('d1', {'null_count': 1}))
        self.assertFalse(mvd.detect('d1', {'null_count': 1}))



from typing import Optional
import logging
import os
import gzip
from pathlib import Path
import onnxruntime

import graphsignal
from graphsignal.proto_utils import parse_semver
from graphsignal.proto import signals_pb2
from graphsignal.profilers.operation_profiler import OperationProfiler
from graphsignal.profilers.profiler_utils import create_log_dir, remove_log_dir, find_and_read

logger = logging.getLogger('graphsignal')


class ONNXRuntimeProfiler(OperationProfiler):
    def __init__(self):
        self._onnx_version = None
        self._onnx_session = None
        self._log_dir = None
        self._ended = False

    def read_info(self, signal):
        if not self._onnx_version:
            self._onnx_version = signals_pb2.SemVer()
            parse_semver(self._onnx_version, onnxruntime.__version__)

        # Framework info
        framework = signal.frameworks.add()
        framework.type = signals_pb2.FrameworkInfo.FrameworkType.ONNX_FRAMEWORK
        framework.version.CopyFrom(self._onnx_version)

    def initialize_options(self, onnx_session_options: onnxruntime.SessionOptions):
        graphsignal._check_configured()

        onnx_session_options.enable_profiling = True

        self._log_dir = '{0}{1}'.format(create_log_dir(), os.path.sep)
        onnx_session_options.profile_file_prefix = self._log_dir

    def set_onnx_session(self, onnx_session):
        self._onnx_session = onnx_session

    def start(self, signal):
        if not self._log_dir:
            raise ValueError('ONNX Runtime profiler is not initialized')

        if not self._onnx_session:
            raise ValueError('ONNX Runtime session is not provided')

        if not self._onnx_session.get_session_options().enable_profiling:
            raise ValueError('ONNX Runtime profiling is not enabled')

        if self._ended:
            return

        logger.debug('Activating ONNX profiler')

        # Profiler info
        signal.agent_info.operation_profiler_type = signals_pb2.AgentInfo.ProfilerType.ONNX_PROFILER

    def stop(self, signal):
        if self._ended:
            return
        self._ended = True

        logger.debug('Deactivating ONNX profiler')

        try:
            trace_file_path = self._onnx_session.end_profiling()
            trace_file_size = os.path.getsize(trace_file_path)
            if trace_file_size > 50 * 1e6:
                raise Exception('Trace file too big: {0}'.format(trace_file_size))

            with open(trace_file_path) as f:
                trace_json = f.read()
                signal.trace_data = gzip.compress(trace_json.encode())
        finally:
            remove_log_dir(self._log_dir)




import logging
import functools

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data.data_profiler import DataProfiler, add_counts

logger = logging.getLogger('graphsignal')


class TorchTensorProfiler(DataProfiler):
    def __init__(self):
        super().__init__()

    def is_instance(self, data):
        torch = self.check_module('torch')
        return torch is not None and torch.is_tensor(data)

    def compute_counts(self, data):
        torch = self.check_module('torch')
        counts = {}
        counts['element_count'] = functools.reduce(lambda x, y: x*y, list(data.size()))
        counts['nan_count'] = int(torch.count_nonzero(torch.isnan(data)))
        counts['inf_count'] = int(torch.count_nonzero(torch.isinf(data)))
        counts['zero_count'] = int(torch.count_nonzero(torch.eq(data, 0)))
        counts['negative_count'] = int(torch.count_nonzero(torch.lt(data, 0)))
        counts['positive_count'] = int(torch.count_nonzero(torch.gt(data, 0)))
        return counts

    def build_stats(self, data):
        counts = self.compute_counts(data)
        data_stats = signals_pb2.DataStats()
        data_stats.data_type = 'torch.Tensor'
        data_stats.shape[:] = list(data.size())
        add_counts(data_stats, counts)
        return data_stats


import logging
import sys
import importlib
import functools

import graphsignal

logger = logging.getLogger('graphsignal')


class DataProfiler:
    def __init__(self):
        self._module = None
        self._is_checked = False

    def check_module(self, name):
        if not self._is_checked:
            self._is_checked = True
            if name in sys.modules:
                try:
                    self._module = importlib.import_module(name)
                except ImportError:
                    logger.error('Error importing {0}'.format(name))            
        return self._module

    def is_instance(self, data):
        raise NotImplementedError()

    def compute_counts(self, data):
        raise NotImplementedError()

    def build_stats(self, data):
        raise NotImplementedError()


def add_counts(data_stats, counts):
    for name, count in counts.items():
        dc = data_stats.data_counts.add()
        dc.name = name
        dc.count = count



import unittest
import logging
import sys
from unittest.mock import patch, Mock
import torch
from google.protobuf.json_format import MessageToJson
import pprint
import time

import graphsignal
from graphsignal.usage.nvml_reader import NvmlReader
from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')


class NvmlReaderTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_read(self):
        signal = signals_pb2.WorkerSignal()

        x = torch.arange(-50, 50, 0.1).view(-1, 1)
        y = -5 * x + 0.1 * torch.randn(x.size())
        model = torch.nn.Linear(1, 1)
        if torch.cuda.is_available():
            x = x.to('cuda:0')
            y = y.to('cuda:0')
            model = model.to('cuda:0')
        criterion = torch.nn.MSELoss()
        optimizer = torch.optim.SGD(model.parameters(), lr=0.1)

        reader = NvmlReader()
        reader.read(signal)

        time.sleep(0.2)

        reader.read(signal)

        #pp = pprint.PrettyPrinter()
        # pp.pprint(MessageToJson(signal))

        if len(signal.device_usage) > 0:
            self.assertTrue(signal.node_usage.num_devices > 0)

            device_usage = signal.device_usage[0]
            self.assertEqual(device_usage.device_type, signals_pb2.DeviceType.GPU)
            self.assertNotEqual(device_usage.device_id, '')
            self.assertNotEqual(device_usage.device_name, '')
            self.assertNotEqual(device_usage.architecture, '')
            self.assertTrue(device_usage.compute_capability.major > 0)
            self.assertTrue(device_usage.mem_total > 0)
            self.assertTrue(device_usage.mem_used > 0)
            self.assertTrue(device_usage.mem_free > 0)
            self.assertTrue(device_usage.gpu_utilization_percent > 0)
            #self.assertTrue(device_usage.mem_access_percent > 0)
            self.assertTrue(device_usage.gpu_temp_c > 0)
            self.assertTrue(device_usage.power_usage_w > 0)
            #self.assertTrue(device_usage.fan_speed_percent > 0)
            self.assertTrue(device_usage.gpu_temp_c > 0)
            self.assertTrue(device_usage.power_usage_w > 0)



import unittest
import logging
import sys
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')


class HuggingFaceSubclassTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    @unittest.skip("long test")
    def test_subclass(self):
        import torch
        from datasets import load_dataset
        raw_datasets = load_dataset("imdb")

        from transformers import AutoTokenizer
        tokenizer = AutoTokenizer.from_pretrained("bert-base-cased", cache_dir='temp')

        def tokenize_function(examples):
            return tokenizer(examples["text"],
                             padding="max_length", truncation=True)

        tokenized_datasets = raw_datasets.map(tokenize_function, batched=True)

        small_train_dataset = tokenized_datasets["train"].shuffle(
            seed=42).select(
            range(10))
        small_eval_dataset = tokenized_datasets["test"].shuffle(
            seed=42).select(
            range(10))

        from transformers import AutoModelForSequenceClassification
        model = AutoModelForSequenceClassification.from_pretrained(
            "distilbert-base-cased", cache_dir='temp', num_labels=2)

        from transformers import TrainingArguments
        training_args = TrainingArguments("test_trainer",
            num_train_epochs=1)
            
        from transformers import Trainer
        from graphsignal.profilers.pytorch import PyTorchProfiler

        profiler = PyTorchProfiler()
        signal = None

        class MyTrainer(Trainer):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
            
            def prediction_step(self, *args, **kwargs):
                global signal
                signal = signals_pb2.WorkerSignal()
                profiler.start(signal)
                ret = super().prediction_step(*args, **kwargs)
                profiler.stop(signal)
                return ret

        trainer = MyTrainer(
            model=model,
            args=training_args,
            train_dataset=small_train_dataset,
            eval_dataset=small_eval_dataset)

        trainer.train()

        trainer.evaluate()

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(signal))

        test_op_stats = None
        for op_stats in signal.op_stats:
            if 'sgemm' in op_stats.op_name:
                test_op_stats = op_stats
                break
        self.assertIsNotNone(test_op_stats)
        self.assertTrue(test_op_stats.count >= 1)
        import torch
        if torch.cuda.is_available():
            self.assertTrue(test_op_stats.total_device_time_us >= 1)
            self.assertTrue(test_op_stats.self_device_time_us >= 1)
        else:
            self.assertTrue(test_op_stats.total_host_time_us >= 1)
            self.assertTrue(test_op_stats.self_host_time_us >= 1)



from typing import Optional
import logging

from tensorflow import keras
from tensorflow.keras.callbacks import Callback

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import parse_semver

logger = logging.getLogger('graphsignal')


class GraphsignalCallback(Callback):
    def __init__(self, endpoint: str, tags: Optional[dict] = None):
        super().__init__()
        self._keras_version = None
        self._tracer = graphsignal.tracer(with_profiler='tensorflow')
        self._trace = None
        self._endpoint = endpoint
        self._tags = tags

    def on_test_begin(self, logs=None):
        self._configure_profiler()

    def on_test_end(self, logs=None):
        graphsignal.upload()

    def on_test_batch_begin(self, batch, logs=None):
        self._start_profiler()

    def on_test_batch_end(self, batch, logs=None):
        self._stop_profiler()

    def on_predict_begin(self, logs=None):
        self._configure_profiler()

    def on_predict_end(self, logs=None):
        graphsignal.upload()

    def on_predict_batch_begin(self, batch, logs=None):
        self._start_profiler()

    def on_predict_batch_end(self, batch, logs=None):
        self._stop_profiler()

    def _configure_profiler(self):
        try:
            self._keras_version = signals_pb2.SemVer()
            parse_semver(self._keras_version, keras.__version__)
        except Exception:
            logger.error('Error configuring Keras profiler', exc_info=True)

    def _start_profiler(self):
        if not self._trace:
            self._trace = graphsignal.tracer(with_profiler='tensorflow').trace(
                endpoint=self._endpoint,
                tags=self._tags)

    def _stop_profiler(self):
        if self._trace:
            if self._trace.is_tracing():
                self._update_profile()
            self._trace.stop()
            self._trace = None

    def _update_profile(self):
        try:
            signal = self._trace._signal

            signal.agent_info.framework_profiler_type = signals_pb2.AgentInfo.ProfilerType.KERAS_PROFILER

            framework = signal.frameworks.add()
            framework.type = signals_pb2.FrameworkInfo.FrameworkType.KERAS_FRAMEWORK
            framework.version.CopyFrom(self._keras_version)
        except Exception as exc:
            self._trace._add_profiler_exception(exc)



import logging
import os
import sys
import time
import socket

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.usage.pynvml import *

logger = logging.getLogger('graphsignal')

class NvmlReader():
    MIN_SAMPLE_READ_INTERVAL_US = int(10 * 1e6)

    def __init__(self):
        self._is_initialized = False
        self._start_us = int(time.time() * 1e6)

    def setup(self):
        if self._is_initialized:
            return
        try:
            nvmlInit()
            self._is_initialized = True
            logger.debug('Initialized NVML')
        except BaseException:
            logger.debug('Error initializing NVML, skipping GPU usage')

    def shutdown(self):
        if not self._is_initialized:
            return
        try:
            nvmlShutdown()
            self._is_initialized = False
        except BaseException:
            logger.error('Error shutting down NVML', exc_info=True)

    def read(self, signal):
        if not self._is_initialized:
            return

        now_us = int(time.time() * 1e6)

        device_count = nvmlDeviceGetCount()

        signal.node_usage.num_devices = device_count

        for idx in range(0, device_count):
            try:
                handle = nvmlDeviceGetHandleByIndex(idx)
            except NVMLError as err:
                log_nvml_error(err)
                continue

            device_usage = signal.device_usage.add()
            device_usage.device_type = signals_pb2.DeviceType.GPU

            try:
                pci_info = nvmlDeviceGetPciInfo(handle)
                device_usage.device_id = pci_info.busId
            except NVMLError as err:
                log_nvml_error(err)

            try:
                device_usage.device_name = nvmlDeviceGetName(handle)
            except NVMLError as err:
                log_nvml_error(err)

            try:
                arch = nvmlDeviceGetArchitecture(handle)
                if arch == NVML_DEVICE_ARCH_KEPLER:
                    device_usage.architecture = 'Kepler'
                elif arch == NVML_DEVICE_ARCH_MAXWELL:
                    device_usage.architecture = 'Maxwell'
                elif arch == NVML_DEVICE_ARCH_PASCAL:
                    device_usage.architecture = 'Pascal'
                elif arch == NVML_DEVICE_ARCH_VOLTA:
                    device_usage.architecture = 'Volta'
                elif arch == NVML_DEVICE_ARCH_TURING:
                    device_usage.architecture = 'Turing'
                elif arch == NVML_DEVICE_ARCH_AMPERE:
                    device_usage.architecture = 'Ampere'
            except NVMLError as err:
                log_nvml_error(err)

            try:
                cc_major, cc_minor = nvmlDeviceGetCudaComputeCapability(handle)
                device_usage.compute_capability.major = cc_major
                device_usage.compute_capability.minor = cc_minor
            except NVMLError as err:
                log_nvml_error(err)

            try:
                mem_info = nvmlDeviceGetMemoryInfo(handle)
                device_usage.mem_total = mem_info.total
                device_usage.mem_used = mem_info.used
                device_usage.mem_free = mem_info.total - mem_info.used
            except NVMLError as err:
                log_nvml_error(err)


            try:
                last_read_us = max(
                    int(self._start_us),
                    now_us - NvmlReader.MIN_SAMPLE_READ_INTERVAL_US)

                sample_value_type, gpu_samples = nvmlDeviceGetSamples(handle, NVML_GPU_UTILIZATION_SAMPLES, last_read_us)
                device_usage.gpu_utilization_percent = _avg_sample_value(sample_value_type, gpu_samples)

                sample_value_type, mem_samples = nvmlDeviceGetSamples(handle, NVML_MEMORY_UTILIZATION_SAMPLES, last_read_us)
                device_usage.mem_access_percent = _avg_sample_value(sample_value_type, mem_samples)
            except NVMLError as err:
                log_nvml_error(err)

            try:
                device_usage.pcie_throughput_tx = nvmlDeviceGetPcieThroughput(
                    handle, NVML_PCIE_UTIL_TX_BYTES)
                device_usage.pcie_throughput_rx = nvmlDeviceGetPcieThroughput(
                    handle, NVML_PCIE_UTIL_RX_BYTES)
            except NVMLError as err:
                log_nvml_error(err)

            try:
                device_usage.gpu_temp_c = nvmlDeviceGetTemperature(
                    handle, NVML_TEMPERATURE_GPU)
            except NVMLError as err:
                log_nvml_error(err)

            try:
                device_usage.power_usage_w = nvmlDeviceGetPowerUsage(
                    handle) / 1000.0
            except NVMLError as err:
                log_nvml_error(err)

            try:
                device_usage.fan_speed_percent = nvmlDeviceGetFanSpeed(
                    handle)
            except NVMLError as err:
                log_nvml_error(err)


def _avg_sample_value(sample_value_type, samples):
    if not samples:
        return 0

    sample_values = []

    if sample_value_type == NVML_VALUE_TYPE_DOUBLE:
        sample_values = [sample.sampleValue.dVal for sample in samples]
    if sample_value_type == NVML_VALUE_TYPE_UNSIGNED_INT:
        sample_values = [sample.sampleValue.uiVal for sample in samples]
    if sample_value_type == NVML_VALUE_TYPE_UNSIGNED_LONG:
        sample_values = [sample.sampleValue.ulVal for sample in samples]
    if sample_value_type == NVML_VALUE_TYPE_UNSIGNED_LONG_LONG:
        sample_values = [sample.sampleValue.ullVal for sample in samples]

    if len(sample_values) > 0:
        return sum(sample_values) / len(sample_values)

    return 0


def log_nvml_error(err):
    if (err.value == NVML_ERROR_NOT_SUPPORTED):
        logger.debug('NVML call not supported', exc_info=True)
    elif (err.value == NVML_ERROR_NOT_FOUND):
        logger.debug('NVML call not found', exc_info=True)
    else:
        logger.error('Error calling NVML', exc_info=True)





class OperationProfiler():
    def read_info(self, signal):
        raise NotImplementedError()

    def start(self, signal):
        raise NotImplementedError()

    def stop(self, signal):
        raise NotImplementedError()



import unittest
import logging
import sys
import time
from unittest.mock import patch, Mock
import numpy as np
import torch

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data.torch_tensor import TorchTensorProfiler

logger = logging.getLogger('graphsignal')


class TorchTensorProfilerTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_is_instance(self):
        profiler = TorchTensorProfiler()
        self.assertTrue(profiler.is_instance(torch.tensor([])))

    def test_compute_counts(self):
        profiler = TorchTensorProfiler()
        self.assertEqual(
            profiler.compute_counts(torch.tensor(np.asarray([[-1, 2.0, 0, np.inf], [1, 0, 0, np.nan]]))), 
            {'element_count': 8,
             'nan_count': 1,
             'inf_count': 1,
             'zero_count': 3,
             'negative_count': 1,
             'positive_count': 3,})

    def test_build_stats(self):
        profiler = TorchTensorProfiler()
        ds = profiler.build_stats(torch.tensor(np.asarray([[1.], [2]])))
        self.assertEqual(ds.data_type, 'torch.Tensor')
        self.assertEqual(ds.shape, [2, 1])



import logging
import os
import sys
import platform
import socket
import re

from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')

version_regexp = re.compile(r'^(\d+)\.?(\d+)?\.?(\d+)?')


def parse_semver(semver_proto, version):
    version_match = version_regexp.match(str(version))
    if version_match is not None:
        groups = version_match.groups()
        if groups[0] is not None:
            semver_proto.major = int(groups[0])
        if groups[1] is not None:
            semver_proto.minor = int(groups[1])
        if groups[2] is not None:
            semver_proto.patch = int(groups[2])


def compare_semver(semver_proto, version):
    semver_int = semver_proto.major * 1e6 + \
        semver_proto.minor * 1e3 + semver_proto.patch
    version_int = version[0] * 1e6 + version[1] * 1e3 + version[2]
    if semver_int < version_int:
        return -1
    if semver_int > version_int:
        return 1
    else:
        return 0



import logging

logger = logging.getLogger('graphsignal')


MISSING_COUNTERS = {
    'null_count',
    'nan_count',
    'inf_count',
    'empty_count'
}

class MissingValueDetector:
    MIN_BASELINE_SIZE = 30

    def __init__(self):
        self._baselines = {}

    def detect(self, data_name, counts):
        has_missing = False
        for counter_name in MISSING_COUNTERS:
            if counter_name in counts:
                if counts[counter_name] > 0:
                    has_missing = True
                    break

        if data_name in self._baselines:
            baseline = self._baselines[data_name]
            baseline['size'] += 1
            if not baseline['has_missing'] and has_missing:
                baseline['has_missing'] = True
                if baseline['size'] > MissingValueDetector.MIN_BASELINE_SIZE:
                    return True
        else:
            self._baselines[data_name] = dict(size=1, has_missing=has_missing)

        return False



from typing import Optional
import logging
import os
import sys
import time
import json
import tensorflow as tf

import graphsignal
from graphsignal.proto_utils import parse_semver, compare_semver
from graphsignal.proto import signals_pb2
from graphsignal.profilers.operation_profiler import OperationProfiler
from graphsignal.profilers.profiler_utils import create_log_dir, remove_log_dir, convert_tensorflow_profile

logger = logging.getLogger('graphsignal')

class TensorFlowProfiler(OperationProfiler):
    def __init__(self):
        self._log_dir = None
        self._tensorflow_version = None

    def read_info(self, signal):
        if not self._tensorflow_version:
            self._tensorflow_version = signals_pb2.SemVer()
            parse_semver(self._tensorflow_version, tf.__version__)
            if compare_semver(self._tensorflow_version, (2, 2, 0)) == -1:
                raise Exception(
                    'TensorFlow profiling is not supported for versions <=2.2')

        # Framework info
        framework = signal.frameworks.add()
        framework.type = signals_pb2.FrameworkInfo.FrameworkType.TENSORFLOW_FRAMEWORK
        framework.version.CopyFrom(self._tensorflow_version)

    def start(self, signal):
        logger.debug('Activating TensorFlow profiler')

        # Profiler info
        signal.agent_info.operation_profiler_type = signals_pb2.AgentInfo.ProfilerType.TENSORFLOW_PROFILER

        try:
            self._log_dir = create_log_dir()

            options = tf.profiler.experimental.ProfilerOptions(
                host_tracer_level=2,
                python_tracer_level=0,
                device_tracer_level=1)
            tf.profiler.experimental.start(self._log_dir, options=options)
        except Exception as e:
            remove_log_dir(self._log_dir)
            raise e

    def stop(self, signal):
        logger.debug('Deactivating TensorFlow profiler')

        try:
            tf.profiler.experimental.stop()

            convert_tensorflow_profile(self._log_dir, signal)
        finally:
           remove_log_dir(self._log_dir)
